import os, shutil, subprocess, glob, torch, torchaudio

REPOSITORY = {
    'name': 'free-spoken-digit-dataset',
    'url': 'https://github.com/Jakobovski/free-spoken-digit-dataset'
}

N_REC = 50

class TorchFSDDGenerator:
    """A :class:`torch:torch.utils.data.Dataset` generator for splits of the Free Spoken Digit Dataset.

    Parameters
    ----------
    version: str
        The version of FSDD to download from `the GitHub repository <https://github.com/Jakobovski/free-spoken-digit-dataset>`_,
        specified as a branch name (defaults to `'master'`) or Git version tag, e.g. `'v1.0.6'`.

        Alternatively, if you already have a local copy of the dataset that you would like to use,
        you can set this argument to `'local'` and provide a path to the folder containing the WAV files, as the ``path`` argument.

    path: str, optional
        If ``version`` is a Git branch name or version tag, then this is the path where the Git repository will be cloned to
        (a new folder will be created at the specified path). If none is specified, then :py:func:`python:os.getcwd` is used.

        If ``version`` is set to `'local'`, then this is the path to the folder containing the WAV audio recordings.

    transforms: callable, optional
        A callable transformation to apply to a 1D :class:`torch:torch.Tensor` of audio samples.

        .. seealso::

            :class:`TorchFSDD`

    load_all: bool
        Whether or not to load the entire dataset into memory.

        .. seealso::

            :class:`TorchFSDD`

    **args: optional
        Arbitrary keyword arguments passed on to :py:func:`torchaudio:torchaudio.load`.
    """
    def __init__(self, version='master', path=None, transforms=None, load_all=False, **args):
        if version == 'local':
            if path is None:
                raise ValueError('Expected path to be a directory containing WAV recordings')
        else:
            path = os.getcwd() if path is None else path
            repo_path = os.path.join(path, REPOSITORY['name'])

            try:
                subprocess.call(['git', '-C', path, 'clone', REPOSITORY['url'], '--branch', version])
                shutil.move(os.path.join(repo_path, 'recordings'), path)
                path = os.path.join(path, 'recordings')
            finally:
                if os.path.isdir(repo_path):
                    shutil.rmtree(repo_path)

        self.path = path
        self.transforms = transforms
        self.load_all = load_all
        self.args = args
        self.all_files = glob.glob(os.path.join(self.path, '*.wav'))

    def full(self):
        """Generates a data set wrapper for the entire data set.

        Returns
        -------
        full_set: :class:`TorchFSDD`
            The :class:`torch:torch.utils.data.Dataset` wrapper for the full data set.
        """
        return TorchFSDD(self.all_files, self.transforms, self.load_all, **self.args)

    def train_test_split(self, test_size=0.1):
        """Generates training and test data set wrappers.

        Parameters
        ----------
        test_size: 0 < float < 1
            Size of the test data set (as a proportion).

        Returns
        -------
        train_set: :class:`TorchFSDD`
            The training set :class:`torch:torch.utils.data.Dataset` wrapper.

        test_set: :class:`TorchFSDD`
            The test set :class:`torch:torch.utils.data.Dataset` wrapper.
        """
        assert 0. < test_size < 1.

        train_files, test_files = [], []
        n_test = int(N_REC * test_size)

        for file in self.all_files:
            file_name, ext = os.path.splitext(os.path.basename(file))
            digit, name, rec_num = file_name.split('_')
            split = test_files if int(rec_num) + 1 <= n_test else train_files
            split.append(file)

        train_set = TorchFSDD(train_files, self.transforms, self.load_all, **self.args)
        test_set = TorchFSDD(test_files, self.transforms, self.load_all, **self.args)
        return train_set, test_set

    def train_val_test_split(self, test_size=0.1, val_size=0.1):
        """Generates training, validation and test data set wrappers.

        Parameters
        ----------
        test_size: 0 < float < 1
            Size of the test data set (as a proportion).

        val_size: 0 < float < 1
            Size of the validation data set (as a proportion).

        Returns
        -------
        train_set: :class:`TorchFSDD`
            The training set :class:`torch:torch.utils.data.Dataset` wrapper.

        val_set: :class:`TorchFSDD`
            The validation set :class:`torch:torch.utils.data.Dataset` wrapper.

        test_set: :class:`TorchFSDD`
            The test set :class:`torch:torch.utils.data.Dataset` wrapper.
        """
        assert 0. < test_size < 1.
        assert 0. < val_size < 1.
        assert test_size + val_size < 1.

        train_files, val_files, test_files = [], [], []
        n_test, n_val = int(N_REC * test_size), int(N_REC * val_size)

        for file in self.all_files:
            file_name, ext = os.path.splitext(os.path.basename(file))
            digit, name, rec_num = file_name.split('_')
            rec_num = int(rec_num) + 1
            if rec_num <= n_test:
                split = test_files
            elif n_test < rec_num <= n_test + n_val:
                split = val_files
            else:
                split = train_files
            split.append(file)

        train_set = TorchFSDD(train_files, self.transforms, self.load_all, **self.args)
        val_set = TorchFSDD(val_files, self.transforms, self.load_all, **self.args)
        test_set = TorchFSDD(test_files, self.transforms, self.load_all, **self.args)
        return train_set, val_set, test_set

class TorchFSDD(torch.utils.data.Dataset):
    """A :class:`torch:torch.utils.data.Dataset` wrapper for specified
    WAV audio recordings of the Free Spoken Digit Dataset.

    .. tip::

        There should rarely be a situation where you have to initialize this class manually,
        unless you are experimenting with specific subsets of the FSDD. You should use :class:`TorchFSDDGenerator`
        to either load the full data set or generate splits for training/validation/testing.

    Parameters
    ----------
    files: list of str
        List of file paths to the WAV audio recordings for the dataset.

    transforms: callable, optional
        A callable transformation to apply to a 1D :class:`torch:torch.Tensor` of audio samples.

        This can be a single transformation, such as the :class:`TrimSilence` transformation included in this package.

        .. code-block:: python

            from torchfsdd import TorchFSDDGenerator, TrimSilence

            fsdd = TorchFSDDGenerator(transforms=TrimSilence(threshold=150))

        It could also be a series of transformations composed together with :class:`torchvision:torchvision.transforms.Compose`.

        .. code-block:: python

            from torchfsdd import TorchFSDDGenerator, TrimSilence
            from torchaudio.transforms import MFCC
            from torchvision.transforms import Compose

            fsdd = TorchFSDDGenerator(transforms=Compose([
                TrimSilence(threshold=100),
                MFCC(sample_rate=8e3, n_mfcc=13)
            ]))

        There are many useful audio transformations in :py:mod:`torchaudio:torchaudio.transforms` such as :class:`torchaudio:torchaudio.transforms.MFCC`.

    load_all: bool
        Whether or not to load the entire dataset into memory.

        This essentially defeats the point of batching, but the dataset is relatively small
        enough that it can comfortably fit into memory and possibly provide some speed-up.

        If this is set to `True`, then the complete set of raw audio recordings and labels
        (for the specified files) can be accessed with ``self.recordings`` and ``self.labels``.

    **args: optional
        Arbitrary keyword arguments passed on to :py:func:`torchaudio:torchaudio.load`.
    """
    def __init__(self, files, transforms=None, load_all=False, **args):
        super().__init__()
        self.files = files
        self.transforms = transforms
        self.args = args

        get_audio = lambda file: torchaudio.load(file, **self.args)[0]
        get_label = lambda file: int(os.path.basename(file)[0])

        if load_all:
            self.recordings, self.labels = [], []
            for file in self.files:
                self.recordings.append(get_audio(file))
                self.labels.append(get_label(file))

            def _load(self, index):
                return self.recordings[index], self.labels[index]
        else:
            def _load(self, index):
                file = self.files[index]
                return get_audio(file), get_label(file)

        setattr(self.__class__, '_load', _load)

    def __len__(self):
        return len(self.files)

    def __getitem__(self, index):
        # Fetch the audio and corresponding label
        x, y = self._load(index)
        x = x.flatten()

        # Transform data if a transformation is given
        if self.transforms is not None:
            x = self.transforms(x)

        return x, y


import os, shutil, pytest, glob, torch
from torchaudio.transforms import MFCC
from torchvision.transforms import Compose
from torchfsdd import TorchFSDD, TorchFSDDGenerator, TrimSilence

def fetch_rec_num(file):
    name = os.path.splitext(os.path.basename(file))[0]
    rec_num = int(name.split('_')[-1])
    return rec_num

def test_generator_local():
    path = 'lib/test/data/v1.0.10'
    fsdd = TorchFSDDGenerator(version='local', path=path)
    with open('lib/test/filenames/v1.0.10') as f:
        true_files = [l.rstrip() for l in f.readlines()]
    read_files = [os.path.basename(file) for file in fsdd.all_files]
    assert len(true_files) == len(read_files) and sorted(true_files) == sorted(read_files)
    assert fsdd.path == path

def test_generator_repo_master():
    path = 'lib/test/data/recordings'
    try:
        fsdd = TorchFSDDGenerator(version='master', path='lib/test/data')
        assert len(fsdd.all_files) > 0
        assert fsdd.path == path
    finally:
        if os.path.isdir(path):
            shutil.rmtree(path)

def test_generator_repo_tag1():
    path = 'lib/test/data/recordings'
    try:
        fsdd = TorchFSDDGenerator(version='v1.0.3', path='lib/test/data')
        with open('lib/test/filenames/v1.0.3') as f:
            true_files = [l.rstrip() for l in f.readlines()]
        read_files = [os.path.basename(file) for file in fsdd.all_files]
        assert len(true_files) == len(read_files) and sorted(true_files) == sorted(read_files)
        assert fsdd.path == path
    finally:
        if os.path.isdir(path):
            shutil.rmtree(path)

def test_generator_repo_tag2():
    path = 'lib/test/data/recordings'
    try:
        fsdd = TorchFSDDGenerator(version='v1.0.10', path='lib/test/data')
        with open('lib/test/filenames/v1.0.10') as f:
            true_files = [l.rstrip() for l in f.readlines()]
        read_files = [os.path.basename(file) for file in fsdd.all_files]
        assert len(true_files) == len(read_files) and sorted(true_files) == sorted(read_files)
        assert fsdd.path == path
    finally:
        if os.path.isdir(path):
            shutil.rmtree(path)

def test_generator_full():
    path = 'lib/test/data/v1.0.10'
    fsdd = TorchFSDDGenerator(version='local', path=path)
    full = fsdd.full()
    with open('lib/test/filenames/v1.0.10') as f:
        true_files = [l.rstrip() for l in f.readlines()]
    read_files = [os.path.basename(file) for file in full.files]
    assert isinstance(full, TorchFSDD)
    assert len(true_files) == len(read_files) and sorted(true_files) == sorted(read_files)
    assert fsdd.path == path

def test_generator_train_test_split_99_1():
    fsdd = TorchFSDDGenerator(version='local', path='lib/test/data/v1.0.10')

    train, test = fsdd.train_test_split(test_size=0.01)
    assert isinstance(train, TorchFSDD)
    assert isinstance(test, TorchFSDD)

    train_nums = {fetch_rec_num(file) for file in train.files}
    assert train_nums == set(range(50))
    assert len(train) == 3000

    test_nums = {fetch_rec_num(file) for file in test.files}
    assert test_nums == set()
    assert len(test) == 0

def test_generator_train_test_split_95_5():
    fsdd = TorchFSDDGenerator(version='local', path='lib/test/data/v1.0.10')

    train, test = fsdd.train_test_split(test_size=0.05)
    assert isinstance(train, TorchFSDD)
    assert isinstance(test, TorchFSDD)

    train_nums = {fetch_rec_num(file) for file in train.files}
    assert train_nums == set(range(2, 50))
    assert len(train) == 2880

    test_nums = {fetch_rec_num(file) for file in test.files}
    assert test_nums == set(range(2))
    assert len(test) == 120

def test_generator_train_test_split_90_10():
    fsdd = TorchFSDDGenerator(version='local', path='lib/test/data/v1.0.10')

    train, test = fsdd.train_test_split(test_size=0.1)
    assert isinstance(train, TorchFSDD)
    assert isinstance(test, TorchFSDD)

    train_nums = {fetch_rec_num(file) for file in train.files}
    assert train_nums == set(range(5, 50))
    assert len(train) == 2700

    test_nums = {fetch_rec_num(file) for file in test.files}
    assert test_nums == set(range(5))
    assert len(test) == 300

def test_generator_train_test_split_50_50():
    fsdd = TorchFSDDGenerator(version='local', path='lib/test/data/v1.0.10')

    train, test = fsdd.train_test_split(test_size=0.5)
    assert isinstance(train, TorchFSDD)
    assert isinstance(test, TorchFSDD)

    train_nums = {fetch_rec_num(file) for file in train.files}
    assert train_nums == set(range(25, 50))
    assert len(train) == 1500

    test_nums = {fetch_rec_num(file) for file in test.files}
    assert test_nums == set(range(0, 25))
    assert len(test) == 1500

def test_generator_train_test_split_10_90():
    fsdd = TorchFSDDGenerator(version='local', path='lib/test/data/v1.0.10')

    train, test = fsdd.train_test_split(test_size=0.9)
    assert isinstance(train, TorchFSDD)
    assert isinstance(test, TorchFSDD)

    train_nums = {fetch_rec_num(file) for file in train.files}
    assert train_nums == set(range(45, 50))
    assert len(train) == 300

    test_nums = {fetch_rec_num(file) for file in test.files}
    assert test_nums == set(range(45))
    assert len(test) == 2700

def test_generator_train_test_split_5_95():
    fsdd = TorchFSDDGenerator(version='local', path='lib/test/data/v1.0.10')

    train, test = fsdd.train_test_split(test_size=0.95)
    assert isinstance(train, TorchFSDD)
    assert isinstance(test, TorchFSDD)

    train_nums = {fetch_rec_num(file) for file in train.files}
    assert train_nums == set(range(47, 50))
    assert len(train) == 180

    test_nums = {fetch_rec_num(file) for file in test.files}
    assert test_nums == set(range(47))
    assert len(test) == 2820

def test_generator_train_test_split_1_99():
    fsdd = TorchFSDDGenerator(version='local', path='lib/test/data/v1.0.10')

    train, test = fsdd.train_test_split(test_size=0.99)
    assert isinstance(train, TorchFSDD)
    assert isinstance(test, TorchFSDD)

    train_nums = {fetch_rec_num(file) for file in train.files}
    assert train_nums == set(range(49, 50))
    assert len(train) == 60

    test_nums = {fetch_rec_num(file) for file in test.files}
    assert test_nums == set(range(49))
    assert len(test) == 2940

def test_generator_train_val_test_split_98_1_1():
    fsdd = TorchFSDDGenerator(version='local', path='lib/test/data/v1.0.10')

    train, val, test = fsdd.train_val_test_split(test_size=0.01, val_size=0.01)
    assert isinstance(train, TorchFSDD)
    assert isinstance(val, TorchFSDD)
    assert isinstance(test, TorchFSDD)

    train_nums = {fetch_rec_num(file) for file in train.files}
    assert train_nums == set(range(50))
    assert len(train) == 3000

    val_nums = {fetch_rec_num(file) for file in val.files}
    assert val_nums == set()
    assert len(val) == 0

    test_nums = {fetch_rec_num(file) for file in test.files}
    assert test_nums == set()
    assert len(test) == 0

def test_generator_train_val_test_split_95_2p5_2p5():
    fsdd = TorchFSDDGenerator(version='local', path='lib/test/data/v1.0.10')

    train, val, test = fsdd.train_val_test_split(test_size=0.025, val_size=0.025)
    assert isinstance(train, TorchFSDD)
    assert isinstance(val, TorchFSDD)
    assert isinstance(test, TorchFSDD)

    train_nums = {fetch_rec_num(file) for file in train.files}
    assert train_nums == set(range(2, 50))
    assert len(train) == 2880

    val_nums = {fetch_rec_num(file) for file in val.files}
    assert val_nums == set(range(1, 2))
    assert len(val) == 60

    test_nums = {fetch_rec_num(file) for file in test.files}
    assert test_nums == set(range(1))
    assert len(test) == 60

def test_generator_train_val_test_split_50_33_17():
    fsdd = TorchFSDDGenerator(version='local', path='lib/test/data/v1.0.10')

    train, val, test = fsdd.train_val_test_split(test_size=0.17, val_size=0.33)
    assert isinstance(train, TorchFSDD)
    assert isinstance(val, TorchFSDD)
    assert isinstance(test, TorchFSDD)

    train_nums = {fetch_rec_num(file) for file in train.files}
    assert train_nums == set(range(24, 50))
    assert len(train) == 1560

    val_nums = {fetch_rec_num(file) for file in val.files}
    assert val_nums == set(range(8, 24))
    assert len(val) == 960

    test_nums = {fetch_rec_num(file) for file in test.files}
    assert test_nums == set(range(8))
    assert len(test) == 480

def test_generator_train_val_test_split_17_50_33():
    fsdd = TorchFSDDGenerator(version='local', path='lib/test/data/v1.0.10')

    train, val, test = fsdd.train_val_test_split(test_size=0.33, val_size=0.50)
    assert isinstance(train, TorchFSDD)
    assert isinstance(val, TorchFSDD)
    assert isinstance(test, TorchFSDD)

    train_nums = {fetch_rec_num(file) for file in train.files}
    assert train_nums == set(range(41, 50))
    assert len(train) == 540

    val_nums = {fetch_rec_num(file) for file in val.files}
    assert val_nums == set(range(16, 41))
    assert len(val) == 1500

    test_nums = {fetch_rec_num(file) for file in test.files}
    assert test_nums == set(range(16))
    assert len(test) == 960

def test_dataset_no_transforms():
    fsdd = TorchFSDD(glob.glob('lib/test/data/v1.0.10/*.wav'))
    x, y = fsdd[0]
    assert isinstance(x, torch.Tensor)
    assert x.ndim == 1
    assert x.min() >= -1
    assert x.max() <= 1
    # assert y == 7

def test_dataset_transforms_single():
    files = glob.glob('lib/test/data/v1.0.10/*.wav')
    x_original, y = TorchFSDD(files)[0]
    x_trans, _ = TorchFSDD(files, transforms=TrimSilence(threshold=0.1))[0]
    assert isinstance(x_trans, torch.Tensor)
    assert x_trans.ndim == 1
    assert x_trans.min() >= -1
    assert x_trans.max() <= 1
    assert len(x_original) != len(x_trans)
    # assert y == 7

def test_dataset_transforms_multiple():
    n_mfcc = 13
    files = glob.glob('lib/test/data/v1.0.10/*.wav')
    x_original, y = TorchFSDD(files)[0]
    x_trans, _ = TorchFSDD(files, transforms=Compose([
        TrimSilence(threshold=0.05),
        MFCC(sample_rate=8e3, n_mfcc=n_mfcc)
    ]))[0]
    assert isinstance(x_trans, torch.Tensor)
    assert x_trans.ndim == 2
    assert len(x_trans) == n_mfcc
    # assert y == 7

def test_dataset_load_all_false():
    files = glob.glob('lib/test/data/v1.0.10/*.wav')
    fsdd = TorchFSDD(files, load_all=False)
    assert not hasattr(fsdd, 'recordings')
    assert not hasattr(fsdd, 'labels')

def test_dataset_load_all_true():
    files = glob.glob('lib/test/data/v1.0.10/*.wav')
    fsdd = TorchFSDD(files, load_all=True)
    assert hasattr(fsdd, 'recordings')
    assert isinstance(fsdd.recordings, list)
    assert isinstance(fsdd.recordings[0], torch.Tensor)
    assert len(fsdd.recordings) == len(files)
    assert hasattr(fsdd, 'labels')
    assert isinstance(fsdd.labels, list)
    assert isinstance(fsdd.labels[0], int)
    assert len(fsdd.labels) == len(files)



import torch
from torchaudio import load
from torchfsdd import TrimSilence

original, sr = load('lib/test/data/sample.wav')
original = original.flatten()

def test_max_threshold():
    trimmed = TrimSilence(threshold=1.)(original)
    assert len(trimmed) == 0

def test_min_threshold():
    trimmed = TrimSilence(threshold=0.)(original)
    assert torch.eq(trimmed, original).all()

def test_normal_threshold():
    trimmed = TrimSilence(threshold=0.1)(original)
    assert len(trimmed) != len(original)
    assert torch.eq(trimmed, original[9:2301]).all()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import re, platform
from pkg_resources import packaging
from setuptools import setup, find_packages
from pathlib import Path

init = Path(__file__).parent / "lib" / "torchfsdd" / "__init__.py"
def load_meta(meta):
    with open(init, "r") as file:
        info = re.search(rf'^__{meta}__\s*=\s*[\'"]([^\'"]*)[\'"]', file.read(), re.MULTILINE).group(1)
        if not info:
            raise RuntimeError(f"Could not load {repr(meta)} metadata")
        return info

with open('README.md', 'r', encoding='utf8') as fh:
    long_description = fh.read()

# Backports for importlib.metadata for Python versions < v3.8
install_requires = []
if packaging.version.parse(platform.python_version()) < packaging.version.parse('3.8'):
    install_requires.append('importlib_metadata')

setup(
    name = load_meta("name"),
    version = load_meta("version"),
    author = load_meta("author"),
    author_email = load_meta("email"),
    description = 'A utility for wrapping the Free Spoken Digit Dataset into PyTorch-ready data set splits.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/eonu/torch-fsdd',
    project_urls = {
        'Documentation': 'https://torchfsdd.readthedocs.io/en/latest',
        'Bug Tracker': 'https://github.com/eonu/torch-fsdd/issues',
        'Source Code': 'https://github.com/eonu/torch-fsdd'
    },
    license = 'MIT',
    package_dir = {'': 'lib'},
    packages = find_packages(where='lib'),
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Natural Language :: English'
    ],
    python_requires = '>=3.6',
    install_requires = install_requires,
    extras_require = {
        'torch': [
            'torch>=1.8+cpu',
            'torchaudio>=0.8+cpu'
        ],
        'dev': [
            'torch>=1.8+cpu',
            'torchaudio>=0.8+cpu',
            'torchvision>=0.8',
            'sphinx',
            'numpydoc',
            'sphinx_rtd_theme',
            'sphinx-autobuild',
            'm2r2',
            'mistune==0.8.4',
            'Jinja2<3.1',
            'sphinx-version-warning',
            'pytest'
        ]
    }
)


class TrimSilence:
    """Removes the silence at the beginning and end of the passed audio data.

    .. warning::
        This transformation assumes that the audio **is** normalized.

    Parameters
    ----------
    threshold: float
        The maximum amount of noise that is considered silence.
    """
    def __init__(self, threshold):
        assert 0. <= threshold <= 1.
        self.threshold = threshold

    def __call__(self, x):
        """Applies the transformation.

        Parameters
        ----------
        x: torch.Tensor
            A one-dimensional tensor of WAV audio samples.

        Returns
        -------
        x: :class:`torch:torch.Tensor`
            The original tensor trimmed for silence.
        """
        start, end = 0, 0

        for i, sample in enumerate(x):
            if abs(sample) > self.threshold:
                start = i
                break

        # Reverse the array for trimming the end
        for i, sample in enumerate(x.flip(dims=(0,))):
            if abs(sample) > self.threshold:
                end = len(x) - i
                break

        return x[start:end]